import os
import numpy as np
import pandas as pd
import duckdb


# Generate sample data
def generate_sample_data(N=10_000_000, seed=42):
    rng = np.random.default_rng(seed)
    D = rng.choice([0, 1], size=(N, 1))
    X = rng.choice(range(20), (N, 2), True)
    Y = D + X @ np.array([1, 2]).reshape(2, 1) + rng.normal(size=(N, 1))
    Y2 = -1 * D + X @ np.array([1, 2]).reshape(2, 1) + rng.normal(size=(N, 1))
    df = pd.DataFrame(
        np.concatenate([Y, Y2, D, X], axis=1), columns=["Y", "Y2", "D", "f1", "f2"]
    )
    return df


def create_duckdb_database(df, db_name="test_dataset.db", table="data"):
    db_path = os.path.abspath(db_name)
    conn = duckdb.connect(db_path)
    try:
        conn.execute(f"DROP TABLE IF EXISTS {table}")
        conn.execute(f"CREATE TABLE {table} AS SELECT * FROM df")
        result = conn.execute(f"SELECT COUNT(*) FROM {table}").fetchone()
        print(f"Created table '{table}' with {result[0]} rows in database: {db_path}")
    finally:
        conn.close()
    return db_path



import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--force-regen",
        action="store_true",
        default=False,
        help="Force regeneration of test data",
    )


@pytest.fixture(scope="session")
def force_regen(request):
    return request.config.getoption("--force-regen")



import numpy as np
import pytest
import os
from duckreg.estimators import DuckRegression
from tests.utils import generate_sample_data, create_duckdb_database
import duckdb
import pandas as pd
import pyfixest as pf


@pytest.fixture(scope="session")
def get_data(force_regen):
    if force_regen:
        return generate_sample_data(1_000_000, seed=42)
    else:
        return generate_sample_data(1_000_000, seed=42)


@pytest.fixture(scope="session")
def database(get_data, force_regen):
    df = get_data
    db_name = "test_dataset.db"
    if force_regen and os.path.exists(db_name):
        os.remove(db_name)
    db_path = create_duckdb_database(df, db_name)
    return db_path


def get_numpy_coefficients(db_path, formula):
    conn = duckdb.connect(db_path)
    df = conn.execute("SELECT * FROM data").df()
    conn.close()

    y = df["Y"].values
    X_cols = [x.strip() for x in formula.split("~")[1].strip().split("+")]
    X = df[X_cols].values
    X = np.column_stack([np.ones(X.shape[0]), X])

    coeffs = np.linalg.inv(X.T @ X) @ X.T @ y
    return coeffs[1:]


def get_pyfixest_estimates(db_path, formula):
    conn = duckdb.connect(db_path)
    df = conn.execute("SELECT * FROM data").df()
    conn.close()
    m_pf = pf.feols(formula, data=df, vcov="hetero")
    return m_pf


@pytest.mark.parametrize(
    "fml",
    [
        "Y ~ D",
        "Y ~ D + f1",
        "Y ~ D + f1 + f2",
    ],
)
def test_fitters(database, fml):
    db_path = database

    m_duck = DuckRegression(
        db_name=db_path,
        table_name="data",
        formula=fml,
        cluster_col="",
        n_bootstraps=0,
        seed=42,
    )
    m_duck.fit()
    m_duck.fit_vcov()
    # nobs
    np.testing.assert_allclose(
        m_duck.df_compressed["count"].sum(), 1_000_000, rtol=1e-4
    ), "Number of observations are not equal"

    results = m_duck.summary()
    compressed_coeffs, compressed_se = (
        results["point_estimate"][1:],
        results["standard_error"][1:],
    )
    uncompressed_coeffs = get_numpy_coefficients(db_path, fml)
    uncompressed_coeffs2 = get_pyfixest_estimates(db_path, fml)

    np.testing.assert_allclose(
        compressed_coeffs, uncompressed_coeffs, rtol=1e-4
    ), f"Coefficients are not equal for formula {fml}"

    np.testing.assert_allclose(
        compressed_coeffs, uncompressed_coeffs2.coef().values[1:], rtol=1e-4
    ), f"Coefficients are not equal to pyfixest version for formula {fml}"

    np.testing.assert_allclose(
        compressed_se, uncompressed_coeffs2.se().values[1:], rtol=1e-4
    ), f"Standard errors are not equal to pyfixest version for formula {fml}"



from abc import ABC, abstractmethod
import duckdb
import numpy as np


######################################################################
class DuckReg(ABC):
    def __init__(
        self,
        db_name: str,
        table_name: str,
        seed: int,
        n_bootstraps: int = 100,
        fitter="numpy",
        keep_connection_open=False,
    ):
        self.db_name = db_name
        self.table_name = table_name
        self.n_bootstraps = n_bootstraps
        self.seed = seed
        self.conn = duckdb.connect(db_name)
        self.rng = np.random.default_rng(seed)
        self.fitter = fitter
        self.keep_connection_open = keep_connection_open

    @abstractmethod
    def prepare_data(self):
        pass

    @abstractmethod
    def compress_data(self):
        pass

    @abstractmethod
    def collect_data(self):
        pass

    @abstractmethod
    def estimate(self):
        pass

    @abstractmethod
    def estimate_feols(self):
        pass

    @abstractmethod
    def bootstrap(self):
        pass

    def fit(self):

        self.prepare_data()
        self.compress_data()

        if self.fitter == "numpy":
            self.point_estimate = self.estimate()
            if self.n_bootstraps > 0:
                self.vcov = self.bootstrap()
            self.conn.close() if not self.keep_connection_open else None
            return None
        elif self.fitter == "feols":
            fit = self.estimate_feols()
            self.point_estimate = fit.coef().values
            if self.n_bootstraps > 0:
                self.vcov = self.bootstrap()
            fit._vcov = self.vcov
            fit.get_inference()
            fit._vcov_type = "NP-Bootstrap"
            fit._vcov_type_detail = "NP-Bootstrap"
            self.conn.close() if not self.keep_connection_open else None
            return fit

        else:
            raise ValueError(
                "Argument 'fitter' must be 'numpy' or 'feols', got {}".format(
                    self.fitter
                )
            )

    def summary(self) -> dict:
        """Summary of regression

        Returns:
            dict
        """
        if self.n_bootstraps > 0:
            return {
                "point_estimate": self.point_estimate,
                "standard_error": np.sqrt(np.diag(self.vcov)),
            }
        return {"point_estimate": self.point_estimate}

    def queries(self) -> dict:
        """Collect all query methods in the class

        Returns:
            dict: Dictionary of query methods
        """
        self._query_names = [x for x in dir(self) if "query" in x]
        self.queries = {
            k: getattr(self, self._query_names[c])
            for c, k in enumerate(self._query_names)
        }
        return self.queries


def wls(X: np.ndarray, y: np.ndarray, n: np.ndarray) -> np.ndarray:
    """Weighted least squares with frequency weights"""
    N = np.sqrt(n)
    N = N.reshape(-1, 1) if N.ndim == 1 else N
    Xn = X * N
    yn = y * N
    betahat = np.linalg.lstsq(Xn, yn, rcond=None)[0]
    return betahat
